from contextlib import contextmanager
from pathlib import Path

from invoke import task

from ..sys import venv
from ..utils import get_symlink_command


PIP_REQUIREMENTS_FILENAME = 'requirements.txt'


def bootstrap(ctx):
    """Initializes a virtual environment for your project."""

    venv.create(
        ctx,
        python_path=ctx.python,
        venv_dir=ctx.paths.remote.project.venv.root)

    with ctx.cd(ctx.paths.remote.project.home):
        install(ctx, package='.', editable=True)
        install(ctx, package='wheel')
        install(ctx, from_req=True)

    symlink_entypoint(ctx)


def symlink_entypoint(ctx):
    """Create a system-wide symlink to a project entrypoint."""
    project_name = ctx.project.name
    ctx.sudo(get_symlink_command(
        '%s/%s' % (ctx.paths.remote.project.venv.bin, project_name),
        '/usr/bin/%s' % project_name))


@contextmanager
def venv_context(ctx):
    """Temporarily switches into virtual environment."""

    with ctx.prefix('. %s/activate' % ctx.paths.remote.project.venv.bin):
        yield


@task
def install(ctx, package='', update=False, from_req=False, editable=False):
    """Installs python package(s) using pip

    :param str|list package:
    :param bool update:
    :param bool from_req:
    :param bool editable: Editable install, aka developer install.

    """
    flags = []

    if update:
        flags.extend(['-U', '--no-cache-dir'])

    editable and flags.append('-e')

    if from_req:
        package = '-r %s' % (Path(ctx.paths.remote.project.home) / PIP_REQUIREMENTS_FILENAME)

    if not isinstance(package, list):
        package = [package]

    with venv_context(ctx):
        ctx.run('pip3 install %s %s' % (' '.join(flags), ' '.join(package)))


@task
def upgrade(ctx, package=''):
    """Upgrades a package.

    If not package name provided upgrades all packages listed in requirements file.

    """
    install(ctx, package=package, update=True, from_req=not package)


@task
def install_vcs(ctx, package, vcs_path):
    """Installs python package(s) using pip from VCS

    :param str|list package: E.g.: sitetree

    :param str vcs_path: E.g.:
        * https://github.com/idlesign/sitetree/@branch
        * git://github.com/idlesign/uwsgiconf@master

    """
    install(ctx, 'git+%s#egg=%s' % (vcs_path, package), editable=True)



from pathlib import Path
from uuid import uuid4

from invoke import task
from patchwork.files import append, exists

from ..utils import echo, cd_sudo


@task
def cat(ctx, path):
    """Outputs file content."""
    ctx.sudo('cat %s' % path)


def append_to_file(ctx, filepath, contents):
    """Add contents to a file.

    :param ctx:
    :param str filepath:
    :param str contents:

    """
    if not exists(ctx, filepath):
        touch(ctx, filepath)
    append(ctx, filepath, contents)


def make_tmp_file(ctx, contents):
    """Makes a temporary file with the given content.
    Returns filepath.

    :param ctx:

    :param str contents:

    :rtype: str

    """
    fpath = '/tmp/wscf_%s' % uuid4()

    append_to_file(ctx, fpath, contents)

    return fpath


def mkdir(ctx, path):
    """Creates a directory."""
    ctx.sudo('mkdir -p %s' % path)


def chmod(ctx, path, mode):
    """Change fs object permissions."""
    ctx.sudo('chmod %s %s' % (mode, path))


def chown(ctx, path, user, group):
    """Sets owner for path contents recursively."""
    ctx.sudo('chown -R %s:%s %s' % (user, group, path))


def touch(ctx, fpath):
    """Creates a file or updates modified date if already exists."""
    ctx.run('touch %s' % fpath)


def setfacl(ctx, path, acl, modify=False):
    """Sets/modifies ACL for a directory/file.

    :param ctx:
    :param path:
    :param acl:
        * u:www-data:--x
        * u::rwX,g::-,o::-,u:idlesign:rwx
        * u::rwx,g::---,o::---,u:idlesign:r-x
    :param modify: Modify or set.

    """
    mode = 'm' if modify else ' --set'

    ctx.sudo('setfacl -R%(mode)s "%(acl)s" %(path)s' % {
        'mode': mode,
        'acl': acl,
        'path': path
    })


def rm(ctx, target, force=True):
    """Removes target file or directory recursively."""
    ctx.sudo('rm -r%s %s' % ('f' if force else '', target))


def gzip_extract(ctx, archive, target_dir=None, do_sudo=False):
    """Extracts gzipped archive into a current directory."""

    target_dir = target_dir or archive.with_name(archive.name.replace('.tar.gz', ''))

    mkdir(ctx, target_dir)

    echo('Extract into %s ...' % target_dir)

    method = ctx.sudo if do_sudo else ctx.run
    method('tar -xzf %s -C %s' % (archive, target_dir))

    return target_dir


def gzip_dir(ctx, src, target_fname, do_sudo=False):
    """GZips a directory."""
    target_fname = str(target_fname)

    arch_ext = '.tar.gz'

    if arch_ext not in target_fname:
        target_fname = '%s%s' % (target_fname, arch_ext)

    echo('Creating %s ...' % target_fname)

    command = 'tar -czf %s *' % target_fname
    command = cd_sudo(src, command)
    method = ctx.sudo if do_sudo else ctx.run

    method(command, warn=True)

    return Path(target_fname)


@task
def tail(ctx, filepath):
    """Tails a file to output."""
    # todo maybe use a more powerful tail from orchestra
    ctx.sudo('tail -f %s' % filepath)


def swap_init(ctx):
    """Creates a swap file."""
    swap_file = '/swapfile'
    ctx.sudo('dd if=/dev/zero of=%s bs=1024 count=524288' % swap_file)
    chmod(ctx, swap_file, 600)
    ctx.sudo('mkswap %s' % swap_file)
    swap_on(ctx)


def swap_on(ctx):
    """Turns on swap."""
    swap_file = '/swapfile'
    ctx.sudo('swapon %s' % swap_file)


def swap_off(ctx):
    """Turns off swap."""
    swap_file = '/swapfile'
    ctx.sudo('swapoff %s' % swap_file)



from os import environ
from pathlib import Path

from patchwork.transfers import rsync as rsync_patchwork


def echo(text):
    """Echoes the text to stdout.

    :param str text:

    """
    print(text)


def link_config(ctx, *, title, name_local, name_remote, dir_remote_confs):

    config_name = name_local
    config_local = Path(ctx.paths.local.configs) / config_name

    if config_local.exists():

        ctx.sudo(get_symlink_command(
            Path(ctx.paths.remote.configs) / config_name,
            dir_remote_confs / name_remote))

        return True

    echo('No %s configuration. Expected: %s' % (title, config_local))

    return False


def rsync(ctx, *args, **kwargs):
    """Ugly workaround for https://github.com/fabric/patchwork/issues/16."""
    ssh_agent = environ.get('SSH_AUTH_SOCK', None)

    if ssh_agent:
        ctx.config.run.env['SSH_AUTH_SOCK'] = ssh_agent

    return rsync_patchwork(ctx, *args, **kwargs)


def cd_sudo(cd, command):
    # https://github.com/pyinvoke/invoke/issues/459
    return 'bash -c "cd %s && %s"' % (cd, command)


def get_symlink_command(src, dest):
    """Returns a symlink command.

    :param src: Source filepath
    :param dest: Link filepath

    """
    return 'ln -sf %s %s' % (src, dest)



from pathlib import Path

from invoke import task

from ..sys import fs as sys_fs
from ..utils import get_symlink_command, rsync, echo


def create_environ_file(ctx, type_marker='production'):
    """Creates environment marker file for a given environment type.

    :param ctx:
    :param str type_marker:

    """
    sys_fs.append_to_file(ctx, ctx.paths.remote.project.state.environ, type_marker)


def symlink_home(ctx):
    """Create a directory with project-related symlinks in user's home."""

    home_linkroot = '~/%s' % ctx.project.name

    sys_fs.mkdir(ctx, home_linkroot)

    dir_map = [
        (ctx.paths.remote.project.home, 'code'),
        (ctx.paths.remote.project.state.root, 'state'),
        (ctx.paths.remote.cache, 'cache'),
    ]

    for dir_, linkname in dir_map:
        ctx.sudo(get_symlink_command(dir_, '%s/%s' % (home_linkroot, linkname)))


def upload_configs(ctx):
    """Uploads project configuration files (under `conf` dir) to remote."""

    source = ctx.paths.local.configs + '/'  # slash is rsync semantic specific

    if Path(source).exists():
        rsync(ctx, source, ctx.paths.remote.configs, delete=True)

    else:
        echo('No configuration files uploaded. Expected: %s' % source)


def create_dir(ctx, path):
    """Creates a directory accessible for project user.

    :param ctx:

    :param str path: Directory path.

    """
    group = ctx.project.group
    sys_fs.mkdir(ctx, path)
    sys_fs.chown(ctx, path, ctx.project.user, group)

    # Grant project group full access.
    sys_fs.setfacl(
        ctx, path, 'g:%(group)s:rwX,d:g:%(group)s:rwX' % {'group': group},
        modify=True)


@task
def cache_init(ctx, drop=False):
    """Initializes cache directory.

    :param ctx:
    :param drop: If True - drops if already exists. Otherwise - empties.

    """
    dir_cache = ctx.paths.remote.cache
    command = '%s*' if drop else '%s/*'
    sys_fs.rm(ctx, command % dir_cache)
    create_dir(ctx, dir_cache)


def bootstrap(ctx):
    """Bootstraps filesystem for the project."""

    project = ctx.paths.remote.project
    dir_state = project.state

    dirs = [
        project.home,
        dir_state.root,
        dir_state.dumps,
        dir_state.static,
        dir_state.media,
    ]

    for dir_ in dirs:
        create_dir(ctx, dir_)

    cache_init(ctx, drop=True)
    create_environ_file(ctx)

    symlink_home(ctx)



from pathlib import Path

from invoke import task

from .fs import create_dir
from ..sys import fs as sys_fs
from ..utils import link_config, echo


@task
def manage(ctx, cmd):
    """Runs Django manage command(s).

    :param str|list cmd:

    """
    if not isinstance(cmd, list):
        cmd = [cmd]

    project_name = ctx.project.name
    for command in cmd:
        ctx.sudo('%s %s' % (project_name, command), pty=True, user=project_name)


def rollout(ctx):
    """Rolls out migrations and statics."""
    migrate(ctx)
    manage(ctx, 'collectstatic --noinput')


@task
def migrate(ctx):
    """Runs Django manage command for project to launch migrations."""
    manage(ctx, 'migrate')


def create_superuser(ctx):
    """Runs Django manage command for project to create a superuser.
    Tries to get e-mail from settings, and username from e-mail.

    """
    command = 'createsuperuser'

    username = ''

    email = ctx.project.email or ''
    if email:
        username = email.partition('@')[0]
        command += ' --email %s --username %s' % (email, username)

    echo('\nCreating Django superuser %s ...' % ('[%s]' % username) if username else '')

    manage(ctx, command)


def bootstrap(ctx):
    """Puts Django production settings file to remote."""

    # Create media and static directories.
    dir_state = ctx.paths.remote.project.state
    create_dir(ctx, dir_state.static)
    create_dir(ctx, dir_state.media)

    link_config(
        ctx,
        title='Django',
        name_local='env_production.py',
        name_remote='env_production.py',
        dir_remote_confs=Path(ctx.paths.remote.project.base) / 'settings'
    )

    migrate(ctx)
    create_superuser(ctx)


def dump(ctx, target_dir):
    """Dumps Django related stuff into a target directory."""

    sys_fs.gzip_dir(
        ctx,
        ctx.paths.remote.project.state.media,
        target_dir,
    )


def restore(ctx, source_dir):
    """Restores Django related stuff from a source directory."""

    sys_fs.gzip_extract(
        ctx,
        archive=source_dir / 'media.tar.gz',
        target_dir=ctx.paths.remote.project.state.media
    )



from invoke import task


BOOTSTRAP_SYSTEM_PACKAGES = [
    'python3-dev',
    'python3-pip',
    'python3-venv',
    'python3-wheel',

    # For source builds.
    'build-essential',
    'libjpeg-dev',  # for Pillow
    'libxml2-dev', 'libxslt1-dev',  # for lxml
    'libpcre3-dev', 'libssl-dev',  # for uWSGI with SSL and routing support

    'git',
    'postgresql', 'libpq-dev',
    'certbot',

    # Utils.
    'acl',
    'mc',
    'htop',
    'net-tools',
    'ncdu',
]


@task
def configure(ctx):
    """Continues configuring using dpkg"""
    ctx.sudo('dpkg --configure -a', env={'DEBIAN_FRONTEND': 'noninteractive'})


def upgrade(ctx):
    """Initiates remote OS upgrade procedure."""
    update()
    ctx.sudo('apt-get upgrade')


@task
def update(ctx):
    """Initiates apt cache update."""
    ctx.sudo('apt-get update')


@task
def install(ctx, packages):
    """Installs packages using apt.

    :param packages:

    """
    if not isinstance(packages, list):
        packages = [packages]

    update(ctx)

    ctx.sudo('apt install -y %s' % ' '.join(packages), env={'DEBIAN_FRONTEND': 'noninteractive'})


def bootstrap(ctx):
    """Bootstraps system by installing required packages."""
    install(ctx, BOOTSTRAP_SYSTEM_PACKAGES)



from pathlib import Path

from fabric.main import Fab, Executor, Config
from invoke.config import merge_dicts


class WebscaffExecutor(Executor):
    """Custom webscaff executor."""


class WebscaffConfig(Config):

    prefix = 'wscaff'

    def check_required(self):
        project = self.project

        for setting_name in ('name', 'repo'):
            if not getattr(project, setting_name):
                raise ValueError("Please provide a value for 'project.%s' setting." % setting_name)

    def contribute_missing(self):

        def init_state_dirs(state_root):

            state = paths.state

            if paths.state.root:
                state_root = Path(paths.state.root)

            else:
                paths.state.root = str(state_root)

            # Set defaults for runtime subdirectories.
            for state_dir_name in state._config.keys():

                if state_dir_name == 'root':
                    continue

                if getattr(state, state_dir_name, None) is None:
                    setattr(state, state_dir_name, str(state_root / state_dir_name))

        project_name = self.project.name

        if not self.project.user:
            self.project.user = project_name

        if not self.project.group:
            self.project.group = self.project.user

        # Remote paths:
        paths = self.paths.remote.project

        if not paths.home:
            paths.home = '%s/%s' % (self.paths.remote.projects, project_name)

        home_remote = Path(paths.home)

        if not paths.base:
            paths.base = str(home_remote / project_name)

        if not self.paths.remote.repo:
            self.paths.remote.repo = paths.home

        if not paths.venv.root:
            paths.venv.root = str(home_remote / 'venv')

        if not paths.venv.bin:
            paths.venv.bin = str(Path(paths.venv.root) / 'bin')

        if not self.paths.remote.cache:
            self.paths.remote.cache = str(Path('/var/cache') / project_name)

        # Initialize remote state dirs.
        init_state_dirs(state_root=Path('/var/lib') / project_name)

        # Local paths:
        paths = self.paths.local.project

        if not paths.base:
            paths.base = str(Path(paths.home) / project_name)

        # Initialize local state dirs.
        init_state_dirs(state_root=Path(paths.home) / 'state')

        # Local and remote configs:
        self.paths.local.configs = str(Path(self.paths.local.project.home) / self.dir_conf)
        self.paths.remote.configs = str(home_remote / self.dir_conf)

    @staticmethod
    def global_defaults():
        defaults = Config.global_defaults()

        ours = {
            'forward_agent': True,  # Transparently access private VCS repos, etc.

            'tasks': {'collection_name': 'wscaffs'},

            'python': 'python3',

            'dir_conf': 'conf',

            'project': {
                'name': None,
                'user': None,
                'group': None,
                'repo': None,  # url
                'domain': '',
                'email': '',
            },

            'paths': {
                'remote': {
                    'temp': '/tmp',
                    'cache': None,
                    'projects': '/srv',
                    'configs': None,
                    'repo': None,

                    'project': {
                        'home': None,
                        'base': None,
                        'state': {
                            'root': None,  # State directory root
                            'dumps': None,  # Dumps (DB, etc.) directory
                            'static': None,  # Django static directory
                            'media': None,  # Django media directory
                            'certbot': None,  # Certbot webroot dir
                            'spool': None,  # uWSGI spooler dir
                            'reloader': None,  # uWSGI touch reload file
                            'maintenance': None,  # uWSGI maintenance trigger file
                            'environ': None,  # Project environ type file
                        },
                        'venv': {
                            'root': None,
                            'bin': None,
                        },
                    },
                },
                'local': {
                    'configs': None,
                    'project': {
                        'home': str(Path.cwd()),
                        'base': None,
                        'state': {
                            'root': None,
                            'dumps': None,
                        }
                    }
                }
            }

        }
        merge_dicts(defaults, ours)

        return defaults


class WebscaffProgram(Fab):

    def execute(self):
        hosts = self.core[0].args.hosts

        if not hosts.value:
            # Load hosts from config if not passed via command -H argument
            remotes = self.config.remotes
            hosts.value = ','.join(remotes) if isinstance(remotes, list) else remotes

        super().execute()

    def parse_collection(self):
        super().parse_collection()
        # Load configuration from working directory,
        # but get tasks from webscaff package.
        self.config.set_project_location(str(Path.cwd()))
        self.config.load_project()

    def update_config(self):
        super().update_config()

        config = self.config
        config.check_required()
        config.contribute_missing()



import os

import pytest
from invoke.context import Context
from invoke.executor import Executor


if not os.getcwd().endswith('tests'):
    # To run both from package root and tests/ subdir.
    os.chdir('tests/')


context_commands = []


class CommandResult:

    def __init__(self, command):
        self.command = command

    stdout = 'tstout'
    ok = True


class MockContext(Context):

    def _run(self, runner, command, **kwargs):
        context_commands.append(command)
        return CommandResult(command)

    def _sudo(self, runner, command, **kwargs):
        context_commands.append(command)
        return CommandResult(command)

    def cd(self, path):
        context_commands.append('cd %s' % path)
        return super().cd(path)

    def put(self, what, where):
        context_commands.append('put %s %s' % (what, where))


class MockExecutor(Executor):

    def expand_calls(self, calls):
        for call in calls:
            call.make_context = lambda config: MockContext(config)
        return super().expand_calls(calls)


@pytest.fixture
def run_command_mock(monkeypatch):

    monkeypatch.chdir('../demo')
    monkeypatch.setattr('webscaff.overrides.WebscaffExecutor', MockExecutor)

    from webscaff.cli import program

    def run_command_mock_(command, *args):
        try:
            program.run(['webscaff', command] + list(args), exit=False)
            commands = list(context_commands)
            return commands
        finally:
            context_commands.clear()

    return run_command_mock_



import pytest


from webscaff.overrides import WebscaffConfig


def test_defaults():

    config = WebscaffConfig()

    with pytest.raises(ValueError):
        config.check_required()

    config.project.name = 'testme'
    config.project.repo = 'https://repome'

    config.check_required()
    config.contribute_missing()


    assert config.forward_agent
    assert config.tasks['collection_name'] == 'wscaffs'
    assert config.python == 'python3'
    assert config.dir_conf == 'conf'

    project = config.project
    assert project.name == 'testme'
    assert project.user == 'testme'
    assert project.group == 'testme'
    assert project.repo == 'https://repome'
    assert project.domain == ''
    assert project.email == ''

    paths = config.paths

    paths_remote = paths.remote
    assert paths_remote.temp == '/tmp'
    assert paths_remote.cache == '/var/cache/testme'
    assert paths_remote.projects == '/srv'
    assert paths_remote.configs == '/srv/testme/conf'
    assert paths_remote.repo == '/srv/testme'

    assert paths_remote.project.home == '/srv/testme'
    assert paths_remote.project.base == '/srv/testme/testme'
    assert paths_remote.project.venv.root == '/srv/testme/venv'
    assert paths_remote.project.venv.bin == '/srv/testme/venv/bin'

    remote_state = paths_remote.project.state
    assert remote_state.static == '/var/lib/testme/static'
    assert remote_state.media == '/var/lib/testme/media'
    assert remote_state.certbot == '/var/lib/testme/certbot'
    assert remote_state.spool == '/var/lib/testme/spool'
    assert remote_state.reloader == '/var/lib/testme/reloader'
    assert remote_state.environ == '/var/lib/testme/environ'

    paths_local = paths.local
    assert paths_local.configs.endswith('/tests/conf')
    assert paths_local.project.home.endswith('/tests')
    assert paths_local.project.base.endswith('/tests/testme')
    assert paths_local.project.state.root.endswith('/tests/state')
    assert paths_local.project.state.dumps.endswith('/tests/state/dumps')



from freezegun import freeze_time


class TestProj:

    def test_dump(self, run_command_mock):

        with freeze_time('2020-01-11T11:55'):

            assert run_command_mock('run.backup') == [
                'mkdir -p /var/lib/mydemo/dumps/2020-01-11T1155-mydemo_dump',
                'bash -c "cd /var/lib/mydemo/media && tar -czf /var/lib/mydemo/dumps/2020-01-11T1155-mydemo_dump/media.tar.gz *"',
                'bash -c "cd /etc/letsencrypt && tar -czf /var/lib/mydemo/dumps/2020-01-11T1155-mydemo_dump/certbot.tar.gz *"',
                'pg_dump -Fc mydemo > /var/lib/mydemo/dumps/2020-01-11T1155-mydemo_dump/db.dump',
                'bash -c "cd /var/lib/mydemo/dumps/2020-01-11T1155-mydemo_dump && tar -czf /var/lib/mydemo/dumps/2020-01-11T1155-mydemo_dump.tar.gz *"',
                'rm -rf /var/lib/mydemo/dumps/2020-01-11T1155-mydemo_dump.tar.gz',
                'rm -rf /var/lib/mydemo/dumps/2020-01-11T1155-mydemo_dump',
            ]

    def test_restore(self, run_command_mock):
        assert run_command_mock('run.restore', '/nonexistent') == []
        assert run_command_mock('run.restore', '/tmp') == [
            'put /tmp /var/lib/mydemo/dumps/tmp',
            'mkdir -p /var/lib/mydemo/dumps/tmp',
            'tar -xzf /var/lib/mydemo/dumps/tmp -C /var/lib/mydemo/dumps/tmp',
            'rm -rf /var/lib/mydemo/dumps/tmp',
            'mkdir -p /var/lib/mydemo/media',
            'tar -xzf /var/lib/mydemo/dumps/tmp/media.tar.gz -C /var/lib/mydemo/media',
            'mkdir -p /etc/letsencrypt',
            'tar -xzf /var/lib/mydemo/dumps/tmp/certbot.tar.gz -C /etc/letsencrypt',
            'dropdb mydemo',
            'createdb mydemo',
            'pg_restore --dbname mydemo /var/lib/mydemo/dumps/tmp/db.dump',
            'rm -rf /var/lib/mydemo/dumps/tmp',
        ]

    def test_bootstrap(self, run_command_mock, monkeypatch):
        monkeypatch.setattr('webscaff.commands.utils.rsync_patchwork', lambda *args, **kwargs: None)

        assert run_command_mock('run.initialize') == [
            'whoami',
            'id -u mydemo',
            'useradd -s /bin/bash mydemo',
            'adduser tstout mydemo',
            'apt-get update',
            'apt install -y python3-dev python3-pip python3-venv python3-wheel '
            'build-essential libjpeg-dev libxml2-dev libxslt1-dev libpcre3-dev libssl-dev '
            'git postgresql libpq-dev certbot acl mc htop net-tools ncdu',
            'mkdir -p /srv/mydemo',
            'chown -R mydemo:mydemo /srv/mydemo',
            'setfacl -Rm "g:mydemo:rwX,d:g:mydemo:rwX" /srv/mydemo',
            'mkdir -p /var/lib/mydemo',
            'chown -R mydemo:mydemo /var/lib/mydemo',
            'setfacl -Rm "g:mydemo:rwX,d:g:mydemo:rwX" /var/lib/mydemo',
            'mkdir -p /var/lib/mydemo/dumps',
            'chown -R mydemo:mydemo /var/lib/mydemo/dumps',
            'setfacl -Rm "g:mydemo:rwX,d:g:mydemo:rwX" /var/lib/mydemo/dumps',
            'mkdir -p /var/lib/mydemo/static',
            'chown -R mydemo:mydemo /var/lib/mydemo/static',
            'setfacl -Rm "g:mydemo:rwX,d:g:mydemo:rwX" /var/lib/mydemo/static',
            'mkdir -p /var/lib/mydemo/media',
            'chown -R mydemo:mydemo /var/lib/mydemo/media',
            'setfacl -Rm "g:mydemo:rwX,d:g:mydemo:rwX" /var/lib/mydemo/media',
            'rm -rf /var/cache/mydemo*',
            'mkdir -p /var/cache/mydemo',
            'chown -R mydemo:mydemo /var/cache/mydemo',
            'setfacl -Rm "g:mydemo:rwX,d:g:mydemo:rwX" /var/cache/mydemo',
            'test -e "$(echo /var/lib/mydemo/environ)"',
            'test -e "$(echo /var/lib/mydemo/environ)"',
            'egrep "^production$" "/var/lib/mydemo/environ"',
            'mkdir -p ~/mydemo',
            'ln -sf /srv/mydemo ~/mydemo/code',
            'ln -sf /var/lib/mydemo ~/mydemo/state',
            'ln -sf /var/cache/mydemo ~/mydemo/cache',
            'test -e "$(echo /srv/mydemo/.git)"',
            'test -e "$(echo /srv/mydemo/venv)"',
            'cd /srv/mydemo',
            'pip3 install -e .',
            'pip3 install  wheel',
            'pip3 install  -r /srv/mydemo/requirements.txt',
            'ln -sf /srv/mydemo/venv/bin/mydemo /usr/bin/mydemo',
            'pg_config --version',
            'createdb mydemo',
            'createuser mydemo',
            'psql -c "GRANT ALL PRIVILEGES ON DATABASE mydemo TO mydemo"',
            'service postgresql restart',
            'mkdir -p /var/lib/mydemo/static',
            'chown -R mydemo:mydemo /var/lib/mydemo/static',
            'setfacl -Rm "g:mydemo:rwX,d:g:mydemo:rwX" /var/lib/mydemo/static',
            'mkdir -p /var/lib/mydemo/media',
            'chown -R mydemo:mydemo /var/lib/mydemo/media',
            'setfacl -Rm "g:mydemo:rwX,d:g:mydemo:rwX" /var/lib/mydemo/media',
            'ln -sf /srv/mydemo/conf/env_production.py /srv/mydemo/mydemo/settings/env_production.py',
            'mydemo migrate',
            'mydemo createsuperuser --email idlesign@some.com --username idlesign',
            'touch /var/lib/mydemo/reloader',
            'mkdir -p /var/lib/mydemo/spool',
            'chown -R mydemo:mydemo /var/lib/mydemo/spool',
            'setfacl -Rm "g:mydemo:rwX,d:g:mydemo:rwX" /var/lib/mydemo/spool',
            'systemctl enable --now /srv/mydemo/conf/mydemo.service',
            'mkdir -p /var/lib/mydemo/certbot',
            'chown -R mydemo:mydemo /var/lib/mydemo/certbot',
            'setfacl -Rm "g:mydemo:rwX,d:g:mydemo:rwX" /var/lib/mydemo/certbot',
            'certbot',
            'ln -sf /srv/mydemo/conf/mydemo-certbot-hook.sh /etc/letsencrypt/renewal-hooks/deploy/mydemo',
            'certbot --agree-tos --no-eff-email --email idlesign@some.com certonly --webroot -d mydemo.here -w /var/lib/mydemo/certbot',
            'setfacl -Rm "g:mydemo:--x" /etc/letsencrypt/archive/',
            'setfacl -Rm "g:mydemo:r-x" /etc/letsencrypt/archive/mydemo.here/',
            'setfacl -Rm "g:mydemo:--x" /etc/letsencrypt/live/',
            'setfacl -Rm "g:mydemo:r-x" /etc/letsencrypt/live/mydemo.here/',
            'shutdown -r now',
        ]


class TestSys:

    def test_info(self, run_command_mock):

        assert run_command_mock('info') == [
            'who', 'uname -a', 'cat /etc/timezone', 'uptime', 'df -h', 'journalctl --disk-usage']

    def test_reboot(self, run_command_mock):

        assert run_command_mock('sys.utils.reboot') == ['shutdown -r now']



from contextlib import contextmanager

from .fs import create_dir
from ..sys import fs


def reload_touch(ctx):
    """Touches a file to initiate uWSGI reload procedure."""
    fs.touch(ctx, ctx.paths.remote.project.state.reloader)


def bootstrap(ctx):
    """Bootstraps uWSGI for the project."""

    # Create touch reload file.
    fs.touch(ctx, ctx.paths.remote.project.state.reloader)
    # Add spooler directory.
    create_dir(ctx, ctx.paths.remote.project.state.spool)


def on_503(ctx):
    """Turns on maintenance mode."""
    fs.touch(ctx, ctx.paths.remote.project.state.maintenance)


def off_503(ctx):
    """Turns off maintenance mode."""
    fs.rm(ctx, ctx.paths.remote.project.state.maintenance)


@contextmanager
def maintenance_mode(ctx):
    """Temporarily turn on maintenance mode."""

    on_503(ctx)

    try:
        yield

    finally:
        off_503(ctx)



from .fs import upload_configs
from .utils import bootstrap, backup, restore, py
from ...utils import collection_from_sub

ns = collection_from_sub(__file__, __name__)
ns.register_task(bootstrap, name='initialize')
ns.register_task(backup)
ns.register_task(py)
ns.register_task(restore)
ns.register_task(upload_configs, name='upconf')



from .fs import append_to_file


def create(ctx, user):
    """Creates a user."""
    if get_id(ctx, user) is None:
        ctx.sudo('useradd -s /bin/bash %s' % user)


def add_sudoer(ctx, user):
    """Adds a user to sudoers."""
    append_to_file(ctx, '/etc/sudoers', "'%s ALL=(ALL:ALL) ALL'" % user)


def add_to_group(ctx, user, group):
    """Adds a user into a group.

    Returns ``True`` if is just added (not already was there).

    :rtype: bool
    """
    result = ctx.sudo('adduser %s %s' % (user, group)).stdout.strip()
    return 'Adding' in result


def get_id(ctx, user):
    """Returns user ID. Might be used to check whether user exists."""
    result = ctx.sudo('id -u %s' % user, warn=True).stdout.strip()
    result = int(result) if result.isdigit() else None

    return result


def whoami(ctx):
    """Returns current user name."""
    return ctx.run('whoami').stdout.strip()



from pathlib import Path

from invoke import task


def bootstrap(ctx):
    """Creates a systemd service file and to enable it."""

    project_name = ctx.project.name
    service_file = Path(ctx.paths.remote.configs) / ('%s.service' % project_name)

    ctx.sudo('systemctl enable --now %s' % service_file)


def restart(ctx):
    """Restarts project service."""
    ctx.sudo('systemctl restart %s' % ctx.project.name)


def status(ctx):
    """Returns project service status."""
    ctx.sudo('systemctl status %s' % ctx.project.name)


@task
def stop(ctx):
    """Stops project service."""
    ctx.sudo('systemctl stop %s' % ctx.project.name)



from invoke import task

from ..utils import echo


@task
def reboot(ctx):
    """Reboots remote immediately."""
    ctx.sudo('shutdown -r now')


def set_locale(ctx, locale='ru_RU'):
    """Generates and sets given UTF-8 locale.

    Default: ru_Ru

    """
    ctx.sudo('locale-gen "%s.UTF-8"' % locale)
    ctx.sudo('dpkg-reconfigure locales')
    ctx.sudo('update-locale LC_ALL=%(locale)s.UTF-8 LANG=%(locale)s.UTF-8' % {'locale': locale})


def info(ctx):
    """Prints out remote system information, including kernel info and timezone."""
    ctx.run('who')
    echo('')
    ctx.run('uname -a')
    echo('')
    ctx.run('cat /etc/timezone')
    echo('')
    ctx.run('uptime')
    echo('')
    ctx.run('df -h')
    echo('')
    ctx.sudo('journalctl --disk-usage')


@task
def shutdown(ctx):
    """Turns the remote off immediately."""
    ctx.sudo('shutdown now')



from pathlib import Path

from patchwork.files import exists

from ..sys import git as sys_git


def bootstrap(ctx):
    """Bootstraps a repository your project."""

    project = ctx.project

    if not exists(ctx, Path(ctx.paths.remote.repo) / '.git'):
        sys_git.clone(
            ctx,
            path_base=ctx.paths.remote.project.home,
            repo_url=project.repo,
            dir_target='.')


def pull(ctx):
    """Pulls a relevant version of code."""
    sys_git.pull(ctx, ctx.paths.remote.repo)



from textwrap import dedent

from invoke import task

from ..sys import pg as sys_pg, fs


def bootstrap(ctx):
    """Bootstraps PostgreSQL for the project."""

    sys_pg.configure(
        ctx,
        project_name=ctx.project.name,
        project_user=ctx.project.user)


def dump(ctx, target_dir):
    """Dumps project Database into a target directory."""
    sys_pg.dump(ctx, ctx.project.name, target_dir=target_dir)


def restore(ctx, source_dir):
    """Restores project Database from a backup directory."""
    sys_pg.restore(ctx, ctx.project.name, source_dir=source_dir)


@task
def psql(ctx, command=None):
    """Launches psql command line utility.

    :param ctx:

    :param command: Command to execute in psql, or a filepath
        to a file containing such commands.

    """
    command = command or ''

    if command:
        command = ' -%s "%s"' % ('f' if '/' in command else 'c', command)

    ctx.sudo('psql%s' % command, user=ctx.project.user)


@task
def sizes(ctx, limit=10):
    """Launches psql command to output top n table sizes (data and indexes).

    :param ctx:
    :param int limit: Show top n tables.

    """
    command = '''
    SELECT
        name AS "Table",
        pg_size_pretty(size_data) AS "Size Data",
        pg_size_pretty(size_idx) AS "Size Indexes",
        pg_size_pretty(size_total) AS "Size Total"

    FROM (

        SELECT
            name,
            pg_table_size(path) AS size_data,
            pg_indexes_size(path) AS size_idx,
            pg_total_relation_size(path) AS size_total

        FROM (
            SELECT
              ('"' || table_schema || '"."' || table_name || '"') AS path,
              (table_schema || '.' || table_name) AS name
            FROM information_schema.tables
        ) AS tables
        ORDER BY size_total DESC

    ) AS pretty_sizes LIMIT %s;
    ''' % limit

    command = fs.make_tmp_file(ctx, dedent(command))

    psql(ctx, command)


@task
def reindex(ctx, table):
    """Launches psql command to reindex given table.

    Useful to reclaim space from bloated indexes.

    :param ctx:

    :param str table: Table name

    """
    psql(ctx, 'REINDEX TABLE %s' % table)



from patchwork.files import exists


def create(ctx, python_path, venv_dir):
    """Creates virtual environment using given Python interpreter path
    if not already created.

    :param ctx:
    :param str python_path: Interpreter (cmd name or full path).
    :param str venv_dir: Directory to create for virtual environment.

    """
    if exists(ctx, venv_dir):
        return

    # Call as module in case `virtualenv` app is not [yet] available.
    ctx.run('%s -m venv %s' % (python_path, venv_dir))



from .run.log import stream
from .run.service import restart, status
from .run.utils import rollout
from .run.uwsgi import reload_touch, on_503, off_503
from .sys.utils import reboot, info
from ..utils import collection_from_sub


ns = collection_from_sub(__file__, __name__)

ns.register_task(info)
ns.register_task(on_503, name='off')
ns.register_task(off_503, name='on')
ns.register_task(reload_touch, name='reload')
ns.register_task(restart)
ns.register_task(rollout)
ns.register_task(status)
ns.register_task(stream, name='log')




def clone(ctx, path_base, repo_url, dir_target):
    """Clones a remote repository.


    .. warning:: Make sure ~/.config is not created beforehand with sudo, otherwise you get:
        unable to access '/home/XXX/.config/git/attributes'

    """
    if 'github' in repo_url:
        # Just to make sure ssh agent forwarding works well.
        ctx.run('ssh-keyscan -t rsa github.com >> ~/.ssh/known_hosts', warn=True)
        ctx.run('ssh -T git@github.com', warn=True)

    with ctx.cd(path_base):
        ctx.run('git clone -v %s %s' % (repo_url, dir_target))


def pull(ctx, path_base):
    """Pulls new data from a remote repository master branch.

    Any local changes will be lost.

    """
    with ctx.cd(path_base):
        ctx.run('git reset --hard')
        ctx.run('git pull origin master')



from pathlib import Path

from . import fs
from ..sys import fs as sys_fs
from ..utils import link_config

LETSENCRYPT_DIR = '/etc/letsencrypt'


def bootstrap(ctx):
    """Bootstraps Certbot for the project."""
    fs.create_dir(ctx, ctx.paths.remote.project.state.certbot)
    project_name = ctx.project.name

    # Prepare directories in /etc/letsencrypt/
    ctx.sudo('certbot', warn=True)

    # Link a deploy hook, to allow project user access to it.
    link_config(
        ctx,
        title='certbot hook',
        name_local=project_name + '-certbot-hook.sh',
        name_remote=project_name,
        dir_remote_confs=Path(f'{LETSENCRYPT_DIR}/renewal-hooks/deploy/')
    )

    get_certificate(ctx)


def get_certificate(ctx):
    """Get certificates from Certbot for HTTPS using webroot plugin.

    :param ctx:

    """
    project = ctx.project
    group = ctx.project.group
    domain = project.domain
    email = project.email or ''
    webroot = ctx.paths.remote.project.state.certbot

    if email:
        email = '--email %s' % email

    command = (
        'certbot --agree-tos --no-eff-email %s certonly --webroot -d %s -w %s' %
        (email, domain, webroot))

    ctx.sudo(command)

    # Set access for
    for realm in ['archive', 'live']:
        sys_fs.setfacl(
            ctx,
            path=f'{LETSENCRYPT_DIR}/{realm}/',
            acl=f'g:{group}:--x', modify=True)

        sys_fs.setfacl(
            ctx,
            path=f'{LETSENCRYPT_DIR}/{realm}/{domain}/',
            acl=f'g:{group}:r-x', modify=True)


def dump(ctx, target_dir):
    """Dumps Certbot related stuff into a target directory."""

    sys_fs.gzip_dir(
        ctx,
        LETSENCRYPT_DIR,
        target_dir,
        do_sudo=True
    )


def restore(ctx, source_dir):
    """Restores Certbot related stuff from a source directory."""

    sys_fs.gzip_extract(
        ctx,
        archive=source_dir / 'certbot.tar.gz',
        target_dir=LETSENCRYPT_DIR,
        do_sudo=True
    )



from functools import partial
from pathlib import Path

from invoke import task

from .fs import tail, append_to_file
from ..utils import link_config, echo


def stop(ctx):
    """Stops PostgreSQL."""
    ctx.sudo('service postgresql stop')


@task
def restart(ctx):
    """Restarts PostgreSQL."""
    ctx.sudo('service postgresql restart')


@task
def reload(ctx):
    """Reloads PostgreSQL."""
    ctx.sudo('service postgresql reload')


def get_version(ctx):
    """Returns a list with PostgreSQL version number."""
    number = ctx.run('pg_config --version').stdout.strip()
    number = number.split(' ')[:2][-1].split('.')
    echo('PostgreSQL version: %s' % number)
    return number


@task
def log_main(ctx):
    """Tails main PostgreSQL log."""
    version = [chunk for chunk in get_version(ctx)[:2] if int(chunk)]  # 9.4, but 10
    tail(ctx, '/var/log/postgresql/postgresql-%s-main.log' % '.'.join(version))


def dump(ctx, db_name, target_dir, binary=True):
    """Dumps DB by name into target directory."""
    sudo_pg = partial(ctx.sudo, user='postgres')

    target_path = Path(target_dir) / 'db.dump'
    fmt = '-Fc' if binary else ''

    sudo_pg('pg_dump %s %s > %s' % (fmt, db_name, target_path))

    return target_path


def restore(ctx, db_name, source_dir):
    """Restores DB from a file."""
    sudo_pg = partial(ctx.sudo, user='postgres', warn=True)

    source_path = Path(source_dir) / 'db.dump'

    echo('Restoring DB dump from %s ...' % source_path)

    sudo_pg('dropdb %s' % db_name)
    sudo_pg('createdb %s' % db_name)
    sudo_pg('pg_restore --dbname %s %s' % (db_name, source_path))

    return source_path


def configure(ctx, project_name, project_user):
    """Configures PostgreSQL for the given project."""

    version_ = '.'.join(get_version(ctx)[:-1])

    path_confs = Path('/etc/postgresql/%s/main/' % version_)
    config_name = 'postgresql.conf'
    target_name = '%s.conf' % project_name

    config_linked = link_config(
        ctx,
        title='granular PG',
        name_local=config_name,
        name_remote=target_name,
        dir_remote_confs=path_confs,
    )

    if config_linked:
        # Append into main config an include line.
        append_to_file(ctx, path_confs / config_name, "include = '%s'" % target_name)

    def create_db_and_user():
        sudo_pg = partial(ctx.sudo, user='postgres')

        sudo_pg('createdb %s' % project_name, warn=True)
        sudo_pg('createuser %s' % project_user)  # No password. Using Unix domain sockets.
        sudo_pg('psql -c "GRANT ALL PRIVILEGES ON DATABASE %s TO %s"' % (project_name, project_user))

    create_db_and_user()

    restart(ctx)



import io
import os
import re

from setuptools import setup, find_packages

import sys

PATH_BASE = os.path.dirname(__file__)


def read_file(fpath):
    """Reads a file within package directories."""
    with io.open(os.path.join(PATH_BASE, fpath)) as f:
        return f.read()


def get_version():
    """Returns version number, without module import (which can lead to ImportError
    if some dependencies are unavailable before install."""
    contents = read_file(os.path.join('webscaff', '__init__.py'))
    version = re.search('VERSION = \(([^)]+)\)', contents)
    version = version.group(1).replace(', ', '.').strip()
    return version


setup(
    name='webscaff',
    version=get_version(),
    url='https://github.com/idlesign/webscaff',

    description='Remote scaffolding and orchestration for web applications',
    long_description=read_file('README.rst'),
    license='BSD 3-Clause License',

    author='Igor `idle sign` Starikov',
    author_email='idlesign@yandex.ru',

    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,

    install_requires=[
        'fabric>=2',
        'patchwork>=1.0.1',
    ],
    setup_requires=[] + (['pytest-runner'] if 'test' in sys.argv else []) + [],

    entry_points={
        'console_scripts': ['webscaff = webscaff.cli:program.run']
    },

    test_suite='tests',

    tests_require=[
        'pytest',
        'freezegun',
    ],

    classifiers=[
        # As in https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: BSD License'
    ],
)





from uuid import uuid4
from datetime import datetime
from os import makedirs
from pathlib import Path

from . import fs, dj, service, uwsgi, venv, git, certbot, pg
from ..sys import usr, apt, utils as sys_utils, fs as sys_fs
from ..utils import echo


def bootstrap(ctx):
    """Initializes a remote for your project."""

    project = ctx.project
    me = usr.whoami(ctx)

    usr.create(ctx, project.user)
    just_added = usr.add_to_group(ctx, me, project.group)

    if just_added:
        echo(' * Initial preparation is done. Please rerun the command to proceed.')
        return

    apt.bootstrap(ctx)
    fs.bootstrap(ctx)
    git.bootstrap(ctx)
    fs.upload_configs(ctx)
    venv.bootstrap(ctx)
    pg.bootstrap(ctx)
    dj.bootstrap(ctx)
    uwsgi.bootstrap(ctx)
    service.bootstrap(ctx)
    certbot.bootstrap(ctx)

    echo('* Done. Reboot now ...')

    sys_utils.reboot(ctx)


def rollout(ctx, upgrade_venv=False):
    """Updates remote files from remote repository and rolls out project.

    :param ctx:
    :param upgrade_venv: Whether to update venv using requirements file.

    """
    git.pull(ctx)

    if upgrade_venv:
        venv.upgrade(ctx)

    dj.rollout(ctx)
    uwsgi.reload_touch(ctx)


def backup(ctx):
    """Creates a project backup."""

    path_dump_local = ctx.paths.local.project.state.dumps

    makedirs(path_dump_local, exist_ok=True)

    path_dumps = Path(ctx.paths.remote.project.state.dumps)
    path_dump = path_dumps / ('%s-%s_dump' % (
        datetime.now().strftime('%Y-%m-%dT%H%M'),
        ctx.project.name
    ))

    # Create a subdirectory in dumps.
    sys_fs.mkdir(ctx, path_dump)

    try:
        dj.dump(ctx, target_dir=path_dump / 'media')
        certbot.dump(ctx, target_dir=path_dump / 'certbot')
        pg.dump(ctx, target_dir=path_dump)

        # Archive everything dumped so far.
        path_dump_arch = sys_fs.gzip_dir(ctx, path_dump, path_dump)

        try:
            # Download it to local machine.
            ctx.get(path_dump_arch, str(Path(path_dump_local) / path_dump_arch.name))

        finally:
            sys_fs.rm(ctx, path_dump_arch)

    finally:
        # Remove dumps subdir.
        sys_fs.rm(ctx, path_dump)


def restore(ctx, backup):
    """Restores project related data from a backup file."""
    backup = Path(backup).absolute()

    if not backup.exists():
        echo("Backup file doesn't exist: %s" % backup)
        return

    path_dumps = Path(ctx.paths.remote.project.state.dumps)
    path_remote_archive = path_dumps / backup.name

    echo('Uploading %s ...' % backup)

    ctx.put('%s' % backup, '%s' % path_remote_archive)
    path_remote = sys_fs.gzip_extract(ctx, path_remote_archive)

    try:
        sys_fs.rm(ctx, path_remote_archive)
        dj.restore(ctx, path_remote)
        certbot.restore(ctx, path_remote)
        pg.restore(ctx, path_remote)

    finally:
        sys_fs.rm(ctx, path_remote)


def py(ctx, filepath):
    """Run given file with Python script on remote."""

    script = Path(filepath).absolute()

    if not script.exists():
        echo("File doesn't exist: %s" % script)
        return

    path_temp = Path(ctx.paths.remote.temp)
    path_remote_script = path_temp / str(uuid4())

    ctx.put('%s' % script, '%s' % path_remote_script)

    try:
        with venv.venv_context(ctx):
            ctx.run('python %s' % path_remote_script, warn=True)

    finally:
        sys_fs.rm(ctx, path_remote_script)



from importlib import import_module
from pathlib import Path

from invoke import Collection, task


class WebscaffCollection(Collection):

    def register_task(self, task_func, name=None):
        self.add_task(task(task_func), name=name)


def collection_from_sub(filepath, name):
    """Task gatherer.

    :param filepath:
    :param name:

    """
    dir_base = Path(filepath).parent
    ns = WebscaffCollection()

    def contribute(items):

        for item in items:
            module = import_module('.%s' % item.stem, name)
            ns.add_collection(module)

    dir_packages = [path.parent for path in dir_base.glob('*/__init__.py')]
    contribute(dir_packages)

    submodules = [path for path in dir_base.glob('*.py') if not path.stem.startswith('_')]
    contribute(submodules)

    return ns



import warnings

from . import VERSION_STR
from .commands import ns
from .overrides import WebscaffProgram, WebscaffExecutor, WebscaffConfig

warnings.filterwarnings('ignore')  # filter out CryptographyDeprecationWarning and others


program = WebscaffProgram(
    name='Webscaff',
    version=VERSION_STR,
    executor_class=WebscaffExecutor,
    config_class=WebscaffConfig,
    namespace=ns,
)
